def buf(x):
    return bytes(x) + bytes(64 - len(x))


FEATURE_KBD_DRIVER = 0x5A
FEATURE_KBD_APP = 0x5D
FEATURE_KBD_ID = FEATURE_KBD_APP

xpad_mode_game = 0x01
xpad_mode_wasd = 0x02
xpad_mode_mouse = 0x03

xpad_cmd_set_mode = 0x01
xpad_cmd_set_mapping = 0x02
xpad_cmd_set_js_dz = 0x04
xpad_cmd_set_tr_dz = 0x05
xpad_cmd_set_vibe_intensity = 0x06
xpad_cmd_check_ready = 0x0A
xpad_cmd_set_calibration = 0x0D
xpad_cmd_set_turbo = 0x0F
xpad_cmd_set_response_curve = 0x13
xpad_cmd_set_adz = 0x18

xpad_axis_xy_left = 0x01
xpad_axis_xy_right = 0x02
xpad_axis_z_left = 0x03
xpad_axis_z_right = 0x04

btn_pair_dpad_u_d = 0x01
btn_pair_dpad_l_r = 0x02
btn_pair_ls_rs = 0x03
btn_pair_lb_rb = 0x04
btn_pair_a_b = 0x05
btn_pair_x_y = 0x06
btn_pair_view_menu = 0x07
btn_pair_m1_m2 = 0x08
btn_pair_lt_rt = 0x09

btn_pair_side_left = 0x00
btn_pair_side_right = 0x01

PAD_A = 0x01
PAD_B = 0x02
PAD_X = 0x03
PAD_Y = 0x04

PAD_LB = 0x05
PAD_RB = 0x06

PAD_LS = 0x07
PAD_RS = 0x08

PAD_DPAD_UP = 0x09
PAD_DPAD_DOWN = 0x0A
PAD_DPAD_LEFT = 0x0B
PAD_DPAD_RIGHT = 0x0C

PAD_VIEW = 0x11
PAD_MENU = 0x12
PAD_XBOX = 0x13

MODE_GAME = buf([FEATURE_KBD_ID, 0xD1, xpad_cmd_set_mode, 0x01, xpad_mode_game])

MODE_MOUSE = buf(
    [FEATURE_KBD_ID, 0xD1, xpad_cmd_set_mode, 0x01, xpad_mode_mouse]
)

REMAP_DPAD_UD = buf(
    [
        FEATURE_KBD_ID,
        0xD1,
        xpad_cmd_set_mapping,
        btn_pair_dpad_u_d,
        0x2C,  # Length, 44
        #
        # Each btn_block is 11 bytes.
        #
        # Four btn_blocks:
        # Button 1
        # Button 1 Secondary
        # Button 2
        # Button 2 Secondary
        #
        # Key Groups
        # 1 = Gamepad
        # 2 = Keyboard
        # 3 = Mouse
        # 4 = Key combo, Keyboard
        # 5 = Media
        #
        # btn_block start
        0x01,  # Key Group
        PAD_DPAD_UP,  # Xbox Keycode
        0x00,  # Keyboard Keycode
        0x00,  # Media Keycode
        0x00,  # Mouse Keycode
        0x00,  # Combo length
        0x00,  # Combo Keycode
        0x00,  # Combo Keycode
        0x00,  # Combo Keycode
        0x00,  # Combo Keycode
        0x00,  # Combo Keycode
        #
        # btn_block start
        0x05,
        0x00,
        0x00,
        0x19,  # MEDIA_SHOW_KEYBOARD
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        # btn_block start
        0x01,
        PAD_DPAD_DOWN,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        # btn_block start
        0x04,
        0x00,
        0x00,
        0x00,
        0x00,
        0x03,  # Length
        0x8C,  # KB_LCTL
        0x88,  # KB_LSHIFT
        0x76,  # KB_ESC
        0x00,
        0x00,
    ]
)

REMAP_DPAD_UD_MOUSE = buf(
    [
        FEATURE_KBD_ID,
        0xD1,
        xpad_cmd_set_mapping,
        btn_pair_dpad_u_d,
        0x2C,  # Length, 44
        # btn_block start
        0x02,
        0x00,
        0x98,  # KB_DOWN_ARROW
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        # btn_block start
        0x05,
        0x00,
        0x00,
        0x19,  # MEDIA_SHOW_KEYBOARD
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        # btn_block start
        0x02,
        0x00,
        0x99,  # KB_UP_ARROW
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        # btn_block start
        0x04,
        0x00,
        0x00,
        0x00,
        0x00,
        0x03,  # Length
        0x8C,  # KB_LCTL
        0x88,  # KB_LSHIFT
        0x76,  # KB_ESC
        0x00,
        0x00,
    ]
)

REMAP_DPAD_LR = buf(
    [
        FEATURE_KBD_ID,
        0xD1,
        xpad_cmd_set_mapping,
        btn_pair_dpad_l_r,
        0x2C,  # Length, 44
        # btn_block start
        0x01,
        PAD_DPAD_LEFT,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        # btn_block start
        0x04,
        0x00,
        0x00,
        0x00,
        0x00,
        0x02,  # Length
        0x82,  # KB_META
        0x23,  # KB_D
        0x00,
        0x00,
        0x00,
        # btn_block start
        0x01,
        PAD_DPAD_RIGHT,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        # btn_block start
        0x04,
        0x00,
        0x00,
        0x00,
        0x00,
        0x02,  # Length
        0x82,  # KB_META
        0x0D,  # KB_TAB
        0x00,
        0x00,
        0x00,
    ]
)

REMAP_DPAD_LR_MOUSE = buf(
    [
        FEATURE_KBD_ID,
        0xD1,
        xpad_cmd_set_mapping,
        btn_pair_dpad_l_r,
        0x2C,  # Length, 44
        # btn_block start
        0x02,
        0x00,
        0x9A,  # KB_LEFT_ARROW
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        # btn_block start
        0x04,
        0x00,
        0x00,
        0x00,
        0x00,
        0x02,  # Length
        0x82,  # KB_META
        0x23,  # KB_D
        0x00,
        0x00,
        0x00,
        # btn_block start
        0x02,
        0x00,
        0x9B,  # KB_RIGHT_ARROW
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        # btn_block start
        0x04,
        0x00,
        0x00,
        0x00,
        0x00,
        0x02,  # Length
        0x82,  # KB_META
        0x0D,  # KB_TAB
        0x00,
        0x00,
        0x00,
    ]
)

REMAP_JOYSTICKS = buf(
    [
        FEATURE_KBD_ID,
        0xD1,
        xpad_cmd_set_mapping,
        btn_pair_ls_rs,
        0x2C,  # Length, 44
        # btn_block start
        0x01,
        PAD_LS,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        # btn_block start
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        # btn_block start
        0x01,
        PAD_RS,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        # btn_block start
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
    ]
)

REMAP_JOYSTICKS_MOUSE = buf(
    [
        FEATURE_KBD_ID,
        0xD1,
        xpad_cmd_set_mapping,
        btn_pair_ls_rs,
        0x2C,  # Length, 44
        # btn_block start
        0x02,
        0x00,
        0x88,  # KB_LSHIFT
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        # btn_block start
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        # btn_block start
        0x03,
        0x00,
        0x00,
        0x00,
        0x01,  # RAT_LCLICK
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        # btn_block start
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
    ]
)

REMAP_SHOULDERS = buf(
    [
        FEATURE_KBD_ID,
        0xD1,
        xpad_cmd_set_mapping,
        btn_pair_lb_rb,
        0x2C,  # Length, 44
        # btn_block start
        0x01,
        PAD_LB,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        # btn_block start
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        # btn_block start
        0x01,
        PAD_RB,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        # btn_block start
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
    ]
)

REMAP_SHOULDERS_MOUSE = buf(
    [
        FEATURE_KBD_ID,
        0xD1,
        xpad_cmd_set_mapping,
        btn_pair_lb_rb,
        0x2C,  # Length, 44
        # btn_block start
        0x02,
        0x00,
        0x0D,  # KB_TAB
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        # btn_block start
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        # btn_block start
        0x03,
        0x00,
        0x00,
        0x00,
        0x01,  # RAT_LCLICK
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        # btn_block start
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
    ]
)

REMAP_AB = buf(
    [
        FEATURE_KBD_ID,
        0xD1,
        xpad_cmd_set_mapping,
        btn_pair_a_b,
        0x2C,  # Length, 44
        # btn_block start
        0x01,
        PAD_A,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        # btn_block start
        0x05,
        0x00,
        0x00,
        0x16,  # MEDIA_SCREENSHOT
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        # btn_block start
        0x01,
        PAD_B,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        # btn_block start
        0x04,
        0x00,
        0x00,
        0x00,
        0x00,
        0x02,  # Length
        0x82,  # KB_META
        0x31,  # KB_N
        0x00,
        0x00,
        0x00,
    ]
)

REMAP_AB_MOUSE = buf(
    [
        FEATURE_KBD_ID,
        0xD1,
        xpad_cmd_set_mapping,
        btn_pair_a_b,
        0x2C,  # Length, 44
        # btn_block start
        0x02,
        0x00,
        0x5A,  # KB_RET
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        # btn_block start
        0x05,
        0x00,
        0x00,
        0x16,  # MEDIA_SCREENSHOT
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        # btn_block start
        0x02,
        0x00,
        0x76,  # KB_ESC
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        # btn_block start
        0x04,
        0x00,
        0x00,
        0x00,
        0x00,
        0x02,  # Length
        0x82,  # KB_META
        0x31,  # KB_N
        0x00,
        0x00,
        0x00,
    ]
)

REMAP_XY = buf(
    [
        FEATURE_KBD_ID,
        0xD1,
        xpad_cmd_set_mapping,
        btn_pair_x_y,
        0x2C,  # Length, 44
        # btn_block start
        0x01,
        PAD_X,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        # btn_block start
        0x04,
        0x00,
        0x00,
        0x00,
        0x00,
        0x02,  # Length
        0x82,  # KB_META
        0x4D,  # KB_P
        0x00,
        0x00,
        0x00,
        # btn_block start
        0x01,
        PAD_Y,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        # btn_block start
        0x05,
        0x00,
        0x00,
        0x1E,  # MEDIA_START_RECORDING
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
    ]
)

REMAP_XY_MOUSE = buf(
    [
        FEATURE_KBD_ID,
        0xD1,
        xpad_cmd_set_mapping,
        btn_pair_x_y,
        0x2C,  # Length, 44
        # btn_block start
        0x02,
        0x00,
        0x97,  # KB_PGDWN
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        # btn_block start
        0x04,
        0x00,
        0x00,
        0x00,
        0x00,
        0x02,  # Length
        0x82,  # KB_META
        0x4D,  # KB_P
        0x00,
        0x00,
        0x00,
        # btn_block start
        0x02,
        0x00,
        0x96,  # KB_PGUP
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        # btn_block start
        0x05,
        0x00,
        0x00,
        0x1E,  # MEDIA_START_RECORDING
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
    ]
)

REMAP_VIEW_MENU = buf(
    [
        FEATURE_KBD_ID,
        0xD1,
        xpad_cmd_set_mapping,
        btn_pair_view_menu,
        0x2C,  # Length, 44
        # btn_block start
        0x01,
        PAD_VIEW,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        # btn_block start
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        # btn_block start
        0x01,
        PAD_MENU,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        # btn_block start
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
    ]
)

REMAP_M1M2_DEFAULT = buf(
    [
        FEATURE_KBD_ID,
        0xD1,
        xpad_cmd_set_mapping,
        btn_pair_m1_m2,
        0x2C,  # Length, 44
        # btn_block start
        0x02,
        0x00,
        0x8E,  # KB_M2
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        # btn_block start
        0x02,
        0x00,
        0x8E,  # KB_M2
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        # btn_block start
        0x02,
        0x00,
        0x8F,  # KB_M1
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        # btn_block start
        0x02,
        0x00,
        0x8F,  # KB_M1
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
    ]
)

REMAP_M1M2_F17F18 = buf(
    [
        FEATURE_KBD_ID,
        0xD1,
        xpad_cmd_set_mapping,
        btn_pair_m1_m2,
        0x2C,  # Length, 44
        # btn_block start
        0x02,
        0x00,
        0x28,  # F17?
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        # btn_block start
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        # btn_block start
        0x02,
        0x00,
        0x30,  # F18?
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        # btn_block start
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
    ]
)

REMAP_TRIGGERS = buf(
    [
        FEATURE_KBD_ID,
        0xD1,
        xpad_cmd_set_mapping,
        btn_pair_lt_rt,
        0x2C,  # Length, 44
        # btn_block start
        0x01,
        0x0D,  # LT
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        # btn_block start
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        # btn_block start
        0x01,
        0x0E,  # RT
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        # btn_block start
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
    ]
)

REMAP_TRIGGERS_MOUSE = buf(
    [
        FEATURE_KBD_ID,
        0xD1,
        xpad_cmd_set_mapping,
        btn_pair_lt_rt,
        0x2C,  # Length, 44
        # btn_block start
        0x04,
        0x00,
        0x00,
        0x00,
        0x00,
        0x02,
        0x88,  # KB_LSHIFT
        0x0D,  # KB_TAB
        0x00,
        0x00,
        0x00,
        # btn_block start
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        # btn_block start
        0x03,
        0x00,
        0x00,
        0x00,
        0x02,  # RAT_RCLICK
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        # btn_block start
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
    ]
)

WAIT_READY = buf([FEATURE_KBD_ID, 0xD1, xpad_cmd_check_ready, 0x01])

COMMIT_RESET = lambda kconf: [
    buf(
        [
            FEATURE_KBD_ID,
            0xD1,
            xpad_cmd_set_turbo,
            0x20,  # Length, 32
            # Turbo buttons go here.
            # Unknown how they are laid out.
        ]
    ),
    buf(
        [
            FEATURE_KBD_ID,
            0xD1,
            xpad_cmd_set_vibe_intensity,
            0x02,  # Length
            kconf.get("vibration_left", kconf.get("vibration", 0x64)),
            kconf.get("vibration_right", kconf.get("vibration", 0x64)),
        ]
    ),
    buf(
        [
            FEATURE_KBD_ID,
            0xD1,
            xpad_cmd_set_js_dz,
            0x04,  # Length
            kconf.get("ls_min", 0x00),  # Left Inner
            kconf.get("ls_max", 0x40),  # Left Outer
            kconf.get("rs_min", 0x00),  # Right Inner
            kconf.get("rs_max", 0x40),  # Right Outer
        ]
    ),
    buf(
        [
            FEATURE_KBD_ID,
            0xD1,
            xpad_cmd_set_tr_dz,
            0x04,  # Length
            kconf.get("lt_min", 0x00),  # Left Inner
            kconf.get("lt_max", 0x40),  # Left Outer
            kconf.get("rt_min", 0x00),  # Right Inner
            kconf.get("rt_max", 0x40),  # Right Outer
        ]
    ),
]

COMMANDS_GAME = lambda kconf: [
    MODE_GAME,
    REMAP_DPAD_LR,
    REMAP_DPAD_UD,
    REMAP_JOYSTICKS,
    REMAP_SHOULDERS,
    REMAP_AB,
    REMAP_XY,
    REMAP_VIEW_MENU,
    REMAP_M1M2_F17F18,
    REMAP_TRIGGERS,
    *COMMIT_RESET(kconf),
]

COMMANDS_MOUSE = lambda kconf: [
    MODE_MOUSE,
    REMAP_DPAD_UD_MOUSE,
    REMAP_DPAD_LR_MOUSE,
    REMAP_JOYSTICKS_MOUSE,
    REMAP_SHOULDERS_MOUSE,
    REMAP_AB_MOUSE,
    REMAP_XY_MOUSE,
    REMAP_VIEW_MENU,
    REMAP_M1M2_F17F18,
    REMAP_TRIGGERS_MOUSE,
    *COMMIT_RESET(kconf),
]

RGB_APPLY = buf([FEATURE_KBD_ID, 0xB4])
RGB_SET = buf([FEATURE_KBD_ID, 0xB5])

RGB_PKEY_INIT = lambda key: [
    # buf([key, 0xB9]), # ?
    buf(
        [
            key,
            0x41,
            0x53,
            0x55,
            0x53,
            0x20,
            0x54,
            0x65,
            0x63,
            0x68,
            0x2E,
            0x49,
            0x6E,
            0x63,
            0x2E,
        ]
    ),
    # buf([key, 0x05, 0x20, 0x31, 0x00, 0x08]), # ?
    # Disable AURA Sync
    # buf([key, 0xB7, 0x00]),
]

RGB_INIT = RGB_PKEY_INIT(FEATURE_KBD_ID)


# RGB on when
# "5a d1 09 01 0f <- val bit"
# 0f: all on
# 00: all off
# 09 (08 + 01): boot/shutdown
# 02: awake
# 04: charging sleep
def config_rgb(boot: bool, charging: bool) -> bytes:
    # Always while awake, users can toggle RGB settings for that
    val = 0x02
    if boot:
        val += 0x09
    if charging:
        val += 0x04
    return buf(
        [
            FEATURE_KBD_ID,
            0xD1,
            0x09,
            0x01,
            val,
        ]
    )


# Calibration
# 5a d0 is the main command group
# For left trigger ally executes:
# 5a d0 06 01 01
# 5a d0 01 0c 00 repeatedly (ally responds with data)
# 5a d0 06 01 02
# 5a d0 03 02 20

# Aura mode
# Init / turn off
# 5a c0 00 01 <- this might fix some issues with leds being stuck
# Turn on (after init)
# 5a bc
# Then commands
# 5ad1080cff0000ff0000ff0000ff0000



import logging
from functools import reduce
from typing import (
    Any,
    Literal,
    Mapping,
    MutableMapping,
    Sequence,
    TypedDict,
    cast,
    Protocol,
)
import time
from copy import copy

from .conf import Config

#
# UI settings
#
logger = logging.getLogger(__name__)


class ButtonSetting(TypedDict):
    """Just a button, emits an event. Used for resets, etc."""

    type: Literal["action"]
    tags: Sequence[str]
    title: str
    hint: str | None


class BooleanSetting(TypedDict):
    """Checkbox container."""

    type: Literal["bool"]
    tags: Sequence[str]
    title: str
    hint: str | None

    default: bool | None


class MultipleSetting(TypedDict):
    """Select one container."""

    type: Literal["multiple"]
    tags: Sequence[str]
    title: str
    hint: str | None

    options: Mapping[str, str]
    default: str | None


class DiscreteSetting(TypedDict):
    """Ordered and fixed numerical options (etc. tdp)."""

    type: Literal["discrete"]
    tags: Sequence[str]
    title: str
    hint: str | None

    options: Sequence[int | float]
    default: int | float | None


class NumericalSetting(TypedDict):
    """Floating numerical option."""

    type: Literal["float"]
    tags: Sequence[str]
    title: str
    hint: str | None
    unit: str | None

    min: float | None
    max: float | None
    smin: int | None
    smax: int | None
    step: int | None

    default: float | None


class IntegerSetting(TypedDict):
    """Floating numerical option."""

    type: Literal["int"]
    tags: Sequence[str]
    title: str
    hint: str | None
    unit: str | None

    min: int | None
    max: int | None
    smin: int | None
    smax: int | None
    step: int | None

    default: int | None


class Color(TypedDict):
    red: int
    green: int
    blue: int


class ColorSetting(TypedDict):
    """RGB color setting."""

    type: Literal["color"]
    tags: Sequence[str]
    title: str
    hint: str | None

    default: Color | None


class DisplaySetting(TypedDict):
    """Shows a text value in the UI."""

    type: Literal["display"]
    tags: Sequence[str]
    title: str
    hint: str | None

    config: Any | None
    default: Any | None


class CustomSetting(TypedDict):
    """Custom plugin setting.

    Can be used for any required custom setting that is not covered by the
    default ones (e.g., fan curves, deadzones).

    The setting type is defined by tags.
    Then, the config variable can be used to supply option specific information
    (e.g., for fan curves how many temperature points are available).

    To validate this setting, each loaded plugin's validate function is called,
    with the tags, config data, and the supplied value."""

    type: Literal["custom"]
    tags: Sequence[str]
    title: str
    hint: str | None

    config: Any | None
    default: Any | None


Setting = (
    ButtonSetting
    | BooleanSetting
    | MultipleSetting
    | DiscreteSetting
    | NumericalSetting
    | IntegerSetting
    | ColorSetting
    | CustomSetting
    | DisplaySetting
)

#
# Containers for settings
#


class Container(TypedDict):
    """Holds a variety of settings."""

    type: Literal["container"]
    tags: Sequence[str]
    title: str
    hint: str | None

    children: MutableMapping[str, "Setting | Container | Mode"]


class Mode(TypedDict):
    """Holds a number of containers, only one of whih can be active at a time."""

    type: Literal["mode"]
    tags: Sequence[str]
    title: str
    hint: str | None

    modes: MutableMapping[str, Container]
    default: str | None


STATE_HEADER = (
    "\n"
    + "# Handheld Daemon State Config\n"
    + "#\n"
    + "# This file contains plugin software-only configuration that will be retained\n"
    + "# across reboots. You may edit this file in lueu of using a frontend.\n"
    + "# This header is on the bottom to make editing easier with e.g., nano.\n"
    + "#\n"
    + "# Parameters that are stored in hardware (TDP, RGB colors, etc) and\n"
    + "# risky parameters that might cause instability and should be reset\n"
    + "# across sessions are not part of this file.\n"
    + "# Use profiles to apply changes to these settings.\n"
    + "#\n"
    + "# Persisted (software) parameters are marked by having a default value.\n"
    + "# Non-persisted/hardware parameters do not have a default value.\n"
    + "#\n"
    + "# This file and comments are autogenerated. Your comments will be discarded\n"
    + "# during configuration changes. Parameters with the value `default` are\n"
    + "# ignored and are meant as a template for you to change them.\n"
    + "#\n"
    + "# - CONFIGURATION PARAMETERS\n"
    + "#"
)

PROFILE_HEADER = (
    "\n"
    + "# Handheld Daemon Profile Config\n"
    + "#\n"
    + "# This file contains the configuration options that will be set when\n"
    + "# applying the profile which shares this file name.\n"
    + "# This header is on the bottom to make editing easier with e.g., nano.\n"
    + "#\n"
    + "# Settings are applied once, when applying the profile, and only the ones\n"
    + "# that are stated change. Therefore, they may drift as the system state changes\n"
    + "# (e.g., using native TDP shortcuts, or controller profile shortcuts).\n"
    + "#\n"
    + "# It is possible to set all supported parameters using profiles, and\n"
    + "# it is encouraged for you to stack profiles together.\n"
    + "#\n"
    + "# For example, you can have TDP only profiles that control the energy budget,\n"
    + "# and controller profiles that switch controller behavior.\n"
    + "# Then, depending on the game, you can apply the appropriate 2 profiles\n"
    + "# together.\n"
    + "#\n"
    + "# This file and comments are autogenerated. Your comments will be discarded\n"
    + "# during configuration changes. Parameters with the value `unset` are\n"
    + "# ignored and are meant to act as a template for you to change them.\n"
    + "#\n"
    + "# - CONFIGURATION PARAMETERS\n"
    + "#"
)


Section = Mapping[str, Container]

HHDSettings = Mapping[str, Section]


def parse(d: Setting | Container | Mode, prev: Sequence[str], out: MutableMapping):
    new_prev = list(prev)
    match d["type"]:
        case "container":
            for k, v in d["children"].items():
                parse(v, new_prev + [k], out)
        case "mode":
            out[".".join(new_prev) + ".mode"] = d.get("default", None)

            for k, v in d["modes"].items():
                parse(v, new_prev + [k], out)
        case other:
            out[".".join(new_prev)] = d.get("default", None)


def parse_defaults(sets: HHDSettings):
    out = {}
    for name, sec in sets.items():
        for cname, cont in sec.items():
            parse(cont, [name, cname], out)
    return out


def pick_tag(tag, default, a, b):
    if not b:
        return a.get(tag, default)
    return b.get(tag, a.get(tag, default))


DEFAULT_TAGS = {
    "type": None,
    "title": "",
    "hint": "",
    "unit": "",
    "tags": [],
    "default": None,
}

TYPE_TAGS = {
    "multiple": {"options": {}},
    "discrete": {"options": []},
    "int": {"min": None, "max": None, "step": None, "unit": None, "smin": None, "smax": None},
    "float": {"min": None, "max": None, "step": None, "unit": None, "smin": None, "smax": None},
    "custom": {"config": None},
}


def merge_reduce(
    a: Setting | Container | Mode, b: Setting | Container | Mode | None = None
):
    s = {}
    for tag, default in DEFAULT_TAGS.items():
        s[tag] = pick_tag(tag, default, a, b)

    for tag, default in TYPE_TAGS.get(a["type"], {}).items():
        s[tag] = pick_tag(tag, default, a, b)

    if b and b.get("type", None) == a.get("type", None):
        match s["type"]:
            case "container":
                new_children = dict(a.get("children", {}))
                for k, v in b.get("children", {}).items():
                    if k in new_children:
                        new_children[k] = merge_reduce(new_children[k], v)  # type: ignore
                    else:
                        new_children[k] = merge_reduce(v)  # type: ignore
                s["children"] = new_children
            case "mode":
                new_children = dict(a.get("modes", {}))
                for k, v in b.get("modes", {}).items():
                    if k in new_children:
                        new_children[k] = merge_reduce(new_children[k], v)  # type: ignore
                    else:
                        new_children[k] = merge_reduce(v)  # type: ignore
                s["modes"] = new_children
    else:
        if a.get("type", None) == "container":
            s["children"] = {
                k: merge_reduce(v) for k, v in a.get("children", {}).items()
            }

        if a.get("type", None) == "mode":
            s["modes"] = {k: merge_reduce(v) for k, v in a.get("modes", {}).items()}
    return s


def merge_reduce_sec(a: Section, b: Section):
    out = {k: cast(Container, merge_reduce(v)) for k, v in a.items()}
    for k, v in b.items():
        if k in out:
            out[k] = cast(Container, merge_reduce(out[k], v))
        else:
            out[k] = cast(Container, merge_reduce(v))

    return out


def merge_reduce_secs(a: HHDSettings, b: HHDSettings):
    out = {k: merge_reduce_sec({}, v) for k, v in a.items()}
    for k, v in b.items():
        out[k] = merge_reduce_sec(out.get(k, {}), v)

    return out


def merge_settings(sets: Sequence[HHDSettings]):
    if not sets:
        return {}
    if len(sets) > 1:
        return reduce(merge_reduce_secs, sets)
    return merge_reduce_secs({}, sets[0])


def generate_desc(s: Setting | Container | Mode):
    desc = f"*{s['title']}*\n"
    if h := s.get("hint", None):
        line = ""
        for token in h.split(" "):
            if len(line) + len(token) > 80:
                desc += f"{line}\n"
                line = ""
            line += f"{token} "
        if line:
            desc += f"{line}\n"

    match s["type"]:
        case "mode":
            desc += f"- modes: [{', '.join(map(str, s['modes']))}]\n"
        case "number":
            desc += f"- numerical: ["
            desc += f"{s['min'] if s.get('min', None) is not None else '-inf'}, "
            desc += f"{s['max'] if s.get('max', None) is not None else '+inf'}]\n"
        case "bool":
            desc += f"- boolean: [False, True]\n"
        case "multiple" | "discrete":
            desc += f"- options: [{', '.join(map(str, s['options']))}]\n"
        case "action":
            desc += f"- action: Set to True to run.\n"

    if (d := s.get("default", None)) is not None:
        desc += f"- default: {d}\n"
    return desc[:-1]


def traverse_desc(set: Setting | Container | Mode, prev: Sequence[str]):
    out = []
    out.append(
        (
            prev,
            generate_desc(set),
            max(len(prev) - 1, 0),
            set["type"] in ("mode", "container"),
        )
    )
    match set["type"]:
        case "container":
            for child_name, child in set["children"].items():
                out.extend(traverse_desc(child, [*prev, child_name]))
        case "mode":
            for mode_name, mode in set["modes"].items():
                out.extend(traverse_desc(mode, [*prev, mode_name]))
    return out


def tranverse_desc_sec(set: HHDSettings):
    out = []
    for sec_name, sec in set.items():
        for cont_name, cnt in sec.items():
            out.extend(traverse_desc(cnt, [sec_name, cont_name]))
    return out


def dump_comment(set: HHDSettings, header: str = STATE_HEADER):
    from hhd import RASTER

    out = "#\n#  "
    out += "\n#  ".join(RASTER.split("\n"))
    out += header
    descs = tranverse_desc_sec(set)
    for i, (path, desc, ofs, is_container) in enumerate(descs):
        out += f"\n# {'│' * max((ofs - 1), 0)}┌> {'.'.join(path)}\n# {'│' * ofs} "
        lines = desc.split("\n")
        out += ("\n# " + "│" * ofs + " ").join(lines[:-1])

        next_ofs = descs[i + 1][2] if i < len(descs) - 1 else 0
        if not is_container:
            next_ofs -= 1
        next_ofs = max(min(next_ofs, ofs), 0)
        out += f"\n# {'│' * next_ofs}{'└' * (ofs - next_ofs)} {lines[-1]}"
        out += f"\n# {'│' * next_ofs}"
    # out += "\n\n"
    return out


def dump_setting(
    set: Container | Mode,
    prev: Sequence[str],
    conf: Config,
    unmark: Literal["unset", "default"] = "default",
):
    """Finds the current settings that are set to a default value and swaps them
    for the value `default`. For settings without a default value (temporary),
    it sets them to None to avoid setting them."""
    match set["type"]:
        case "container":
            out = {}
            for child_name, child in set["children"].items():
                match child["type"]:
                    case "container" | "mode":
                        s = dump_setting(child, [*prev, child_name], conf, unmark)
                        if s:
                            out[child_name] = s
                    case _:
                        m = conf.get([*prev, child_name], None)
                        # Skip writing default values
                        default = child.get("default", None)
                        if default is None and unmark != "unset":
                            out[child_name] = None
                        elif m is None:
                            out[child_name] = unmark
                        elif default == m and unmark == "default":
                            out[child_name] = unmark
            return out
        case "mode":
            out = {}
            m = conf.get([*prev, "mode"], None)
            # Skip writing default values
            default = set.get("default", None)
            if default is None and unmark != "unset":
                out["mode"] = None
            elif m is None:
                out["mode"] = unmark
            elif default == m and unmark == "default":
                out["mode"] = unmark

            for mode_name, mode in set["modes"].items():
                s = dump_setting(mode, [*prev, mode_name], conf, unmark)
                if s:
                    out[mode_name] = s
            return out


def merge_dicts(a: Mapping | Any, b: Mapping | Any):
    if isinstance(a, Mapping) and isinstance(b, Mapping):
        out = dict(a)
        for k, v in b.items():
            out[k] = merge_dicts(out.get(k, None), v)
    elif isinstance(b, Mapping):
        out = {}
        for k, v in b.items():
            out[k] = merge_dicts(None, v)
    else:
        return b

    for k in list(out.keys()):
        if out[k] is None:
            del out[k]
    if not out:
        return None
    return out


def dump_settings(
    set: HHDSettings, conf: Config, unmark: Literal["unset", "default"] = "default"
):
    """Fixes default values for settings in set, drops settings without a default value,
    and retains the rest of the configuration, to not mess with plugins that
    were not loaded."""
    out: dict = {"version": get_settings_hash(set)}
    for sec_name, sec in set.items():
        out[sec_name] = {}
        for cont_name, cnt in sec.items():
            s = dump_setting(cnt, [sec_name, cont_name], conf, unmark)
            if s:
                out[sec_name][cont_name] = s

    # Merge dicts to maintain settings for plugins that did not run
    return merge_dicts({"version": None, **cast(Mapping, conf.conf)}, out)


def save_state_yaml(fn: str, set: HHDSettings, conf: Config, shash=None):
    import yaml

    if shash is None:
        shash = get_settings_hash(set)
    if conf.get("version", None) == shash and not conf.updated:
        return False

    conf["version"] = shash
    with open(fn, "w") as f:
        yaml.safe_dump(dump_settings(set, conf, "default"), f, sort_keys=False)
        f.write("\n")
        f.write(dump_comment(set, STATE_HEADER))

    return True


def save_blacklist_yaml(fn: str, avail: Sequence[str], blacklist: Sequence[str]):
    import yaml

    with open(fn, "w") as f:
        f.write(
            (
                ""
                + "# \n"
                + "# Plugin blacklist\n"
                + "# The plugin providers under blacklist will not run.\n"
                + "# \n"
                + "# Warning: this file is read only on startup.\n"
                + "# `sudo systemctl restart hhd@$(whoami)`\n"
                + "# \n"
                + "# Available providers:\n"
                + f"# [{', '.join(avail)}]\n\n"
            )
        )
        yaml.safe_dump({"blacklist": blacklist}, f, width=85, sort_keys=False)

    return True


def load_blacklist_yaml(fn: str):
    import yaml

    try:
        with open(fn, "r") as f:
            return yaml.safe_load(f)["blacklist"]
    except Exception as e:
        logger.warning(f"Plugin blacklist not found, using default (empty).")
        return ["myplugin1"]


def save_profile_yaml(
    fn: str, set: HHDSettings, conf: Config | None = None, shash=None
):
    import yaml

    if shash is None:
        shash = get_settings_hash(set)
    if conf is None:
        conf = Config({})
    elif conf.get("version", None) == shash and not conf.updated:
        return False

    conf["version"] = shash
    with open(fn, "w") as f:
        yaml.safe_dump(dump_settings(set, conf, "unset"), f, width=85, sort_keys=False)
        f.write("\n")
        f.write(dump_comment(set, PROFILE_HEADER))
    return True


def strip_defaults(c):
    if c == "default" or c == "unset":
        return None
    if not isinstance(c, Mapping):
        return c

    out = {}
    for k, v in c.items():
        l = strip_defaults(v)
        if l is not None:
            out[k] = l

    if not out:
        return None
    return out


def get_default_state(set: HHDSettings):
    return Config(parse_defaults(set))


def load_state_yaml(fn: str, set: HHDSettings):
    import yaml

    defaults = parse_defaults(set)
    try:
        with open(fn, "r") as f:
            state = cast(Mapping, strip_defaults(yaml.safe_load(f)) or {})
    except FileNotFoundError:
        logger.warning(f"State file not found. Searched location:\n{fn}")
        return None
    except yaml.YAMLError:
        logger.warning(f"State file is invalid. Searched location:\n{fn}")
        return None

    return Config([defaults, state])


def load_profile_yaml(fn: str):
    import yaml

    try:
        with open(fn, "r") as f:
            state = cast(Mapping, strip_defaults(yaml.safe_load(f)) or {})
    except FileNotFoundError:
        logger.warning(
            f"Profile file not found, using defaults. Searched location:\n{fn}"
        )
        return None
    except yaml.YAMLError:
        logger.warning(
            f"Profile file is invalid, skipping loading. Searched location:\n{fn}"
        )
        return None

    return Config([state])


def get_settings_hash(set: HHDSettings):
    import hashlib, json

    return hashlib.md5(json.dumps(set).encode()).hexdigest()[:8]


def unravel(d: Setting | Container | Mode, prev: Sequence[str], out: MutableMapping):
    new_prev = list(prev)
    match d["type"]:
        case "container":
            for k, v in d["children"].items():
                unravel(v, new_prev + [k], out)
        case "mode":
            out[".".join(new_prev) + ".mode"] = d

            for k, v in d["modes"].items():
                unravel(v, new_prev + [k], out)
        case _:
            out[".".join(new_prev)] = d


def unravel_options(settings: HHDSettings):
    options: Mapping[str, Setting | Mode] = {}
    for name, sec in settings.items():
        for cname, cont in sec.items():
            unravel(cont, [name, cname], options)

    return options


class Validator(Protocol):
    def __call__(self, tags: Sequence[str], config: Any, value: Any) -> bool:
        return False


def validate_config(
    conf: Config, settings: HHDSettings, validator: Validator, use_defaults: bool = True
):
    options = unravel_options(settings)

    for k, d in options.items():
        v = conf.get(k, None)
        if d["type"] == "action":
            default = False
        else:
            default = d["default"]
        if v is None:
            if use_defaults and default is not None:
                conf[k] = default
            continue

        match d["type"]:
            case "mode":
                if v not in d["modes"]:
                    if use_defaults:
                        conf[k] = default
                    else:
                        del conf[k]
            case "bool" | "action":
                if v not in (False, True):
                    conf[k] = bool(v)
            case "multiple" | "discrete":
                if v not in d["options"]:
                    if use_defaults:
                        conf[k] = default
                    else:
                        del conf[k]
            case "int" | "integer":
                if not isinstance(v, int):
                    conf[k] = int(v)
                if v < d["min"]:
                    conf[k] = d["min"]
                if v > d["max"]:
                    conf[k] = d["max"]
            case "float":
                if not isinstance(v, float):
                    conf[k] = float(v)
                if v < d["min"]:
                    conf[k] = d["min"]
                if v > d["max"]:
                    conf[k] = d["max"]
            case "color":
                invalid = False

                if not isinstance(v, Mapping):
                    invalid = True
                else:
                    for c in ("red", "green", "blue"):
                        if c not in v:
                            invalid = True
                        elif not (0 <= v[c] < 256):
                            invalid = True

                if invalid:
                    if use_defaults:
                        conf[k] = default
                    else:
                        del conf[k]
            case "custom":
                if not validator(d["tags"], d["config"], v):
                    if use_defaults:
                        conf[k] = default
                    else:
                        del conf[k]



from hhd.controller import Axis, Button, Configuration
from hhd.controller.physical.evdev import B, to_map
from hhd.plugins import gen_gyro_state

DEFAULT_MAPPINGS: dict[str, tuple[Axis, str | None, float, float | None]] = {
    "accel_x": ("accel_z", "accel", 1, None),
    "accel_y": ("accel_x", "accel", -1, None),
    "accel_z": ("accel_y", "accel", -1, None),
    "anglvel_x": ("gyro_z", "anglvel", 1, None),
    "anglvel_y": ("gyro_x", "anglvel", -1, None),
    "anglvel_z": ("gyro_y", "anglvel", -1, None),
    "timestamp": ("imu_ts", None, 1, None),
}

BTN_MAPPINGS: dict[int, Button] = {
    # Volume buttons come from the same keyboard
    B("KEY_VOLUMEUP"): "key_volumeup",
    B("KEY_VOLUMEDOWN"): "key_volumedown",
    #
    # Loki
    #
    B("KEY_T"): "share",  # T + LCTRL + LSHFT + LALT
}

AMBERNIC_MAPPINGS: dict[int, str] = {
    B("KEY_LEFTMETA"): "share",
    B("KEY_G"): "mode",
}

MSI_CLAW_MAPPINGS = {
    B("KEY_VOLUMEUP"): "key_volumeup",
    B("KEY_VOLUMEDOWN"): "key_volumedown",
    B("KEY_F15"): "mode",
    B("KEY_F16"): "share",
}

AYANEO_DEFAULT_MAPPINGS: dict[str, tuple[Axis, str | None, float, float | None]] = {
    "accel_x": ("accel_z", "accel", 1, None),
    "accel_y": ("accel_x", "accel", -1, None),
    "accel_z": ("accel_y", "accel", -1, None),
    "anglvel_x": ("gyro_z", "anglvel", 1, None),
    "anglvel_y": ("gyro_x", "anglvel", -1, None),
    "anglvel_z": ("gyro_y", "anglvel", -1, None),
    "timestamp": ("imu_ts", None, 1, None),
}

AYANEO_AIR_PLUS_MAPPINGS: dict[str, tuple[Axis, str | None, float, float | None]] = {
    "accel_x": ("accel_z", "accel", -1, None),
    "accel_y": ("accel_x", "accel", -1, None),
    "accel_z": ("accel_y", "accel", 1, None),
    "anglvel_x": ("gyro_z", "anglvel", -1, None),
    "anglvel_y": ("gyro_x", "anglvel", -1, None),
    "anglvel_z": ("gyro_y", "anglvel", 1, None),
    "timestamp": ("imu_ts", None, 1, None),
}

AYANEO_BTN_MAPPINGS: dict[int, str] = {
    # Volume buttons come from the same keyboard
    B("KEY_VOLUMEUP"): "key_volumeup",
    B("KEY_VOLUMEDOWN"): "key_volumedown",
    # Air Plus mappings
    B("KEY_F17"): "mode",  # Big Button
    B("KEY_D"): "share",  # Small Button
    B("KEY_F15"): "extra_l1",  # LC Button
    B("KEY_F16"): "extra_r1",  # RC Button
    # NEXT mappings
    B(
        "KEY_F12"
    ): "mode",  # Big Button NEXT:[[96, 105, 133], [88, 97, 125]] ; Air [88, 97, 125]
    # B("KEY_D"): "share", # Small Button [[40, 133], [32, 125]]
    # 2021 Mappings
    B("KEY_DELETE"): "share",  # TM Button [97,100,111]
    B("KEY_ESC"): "extra_l1",  # ESC Button [1]
    B("KEY_O"): "extra_r1",  # KB Button [97, 24, 125]
    # B("KEY_LEFTMETA"): "extra_r1", # Win Button [125], Conflict with KB Button
    # Air mappings
    B("KEY_F11"): "extra_l1",  # LC Button [87, 97, 125] F11 + LCTRL + LMETA
    B("KEY_F10"): "extra_r1",  # Rc Button [68, 97, 125] F10 + LCTRL + LMETA
}

AYA_DEFAULT_CONF = {
    "hrtimer": True,
    "btn_mapping": AYANEO_BTN_MAPPINGS,
    "mapping": AYANEO_DEFAULT_MAPPINGS,
}
ONEX_DEFAULT_CONF = {
    "hrtimer": True,
}

CONFS = {
    # Ayn
    "Loki MiniPro": {
        "name": "Loki MiniPro",
        "hrtimer": True,
        "mapping": gen_gyro_state("x", False, "z", False, "y", True),
        "extra_buttons": "none",
    },
    "Loki Max": {
        "name": "Loki Max",
        "hrtimer": True,
        "mapping": gen_gyro_state("x", False, "z", False, "y", True),
        "extra_buttons": "none",
    },
    "Loki Zero": {
        "name": "Loki Zero",
        "hrtimer": True,
        "mapping": gen_gyro_state("x", False, "z", False, "y", True),
        "extra_buttons": "none",
    },
    # Ayaneo
    "AIR Plus": {
        "name": "AYANEO AIR Plus",
        **AYA_DEFAULT_CONF,
        "mapping": AYANEO_AIR_PLUS_MAPPINGS,
    },
    "AIR 1S": {"name": "AIR 1S", **AYA_DEFAULT_CONF},
    "AIR 1S Limited": {"name": "AIR 1S Limited", **AYA_DEFAULT_CONF},
    "AYANEO 2": {"name": "AYANEO 2", **AYA_DEFAULT_CONF},
    "AYANEO 2S": {"name": "AYANEO S2", **AYA_DEFAULT_CONF},
    "GEEK": {"name": "AYANEO GEEK", **AYA_DEFAULT_CONF},
    "GEEK 1S": {"name": "AYANEO GEEK 1S", **AYA_DEFAULT_CONF},
    "AIR": {"name": "AYANEO AIR", **AYA_DEFAULT_CONF},
    "AIR Pro": {"name": "AYANEO AIR Pro", **AYA_DEFAULT_CONF},
    "NEXT Advance": {"name": "AYANEO NEXT Advance", **AYA_DEFAULT_CONF},
    "NEXT Lite": {"name": "AYANEO NEXT Lite", **AYA_DEFAULT_CONF},
    "NEXT Pro": {"name": "AYANEO NEXT Pro", **AYA_DEFAULT_CONF},
    "NEXT": {"name": "AYANEO NEXT", **AYA_DEFAULT_CONF},
    "SLIDE": {
        "name": "AYANEO SLIDE",
        **AYA_DEFAULT_CONF,
        "mapping": gen_gyro_state("z", False, "x", False, "y", False),
    },
    "AYA NEO FOUNDER": {"name": "AYA NEO FOUNDER", **AYA_DEFAULT_CONF},
    "AYA NEO 2021": {"name": "AYA NEO 2021", **AYA_DEFAULT_CONF},
    "AYANEO 2021": {"name": "AYANEO 2021", **AYA_DEFAULT_CONF},
    "AYANEO 2021 Pro": {"name": "AYANEO 2021 Pro", **AYA_DEFAULT_CONF},
    "AYANEO 2021 Pro Retro Power": {
        "name": "AYANEO 2021 Pro Retro Power",
        **AYA_DEFAULT_CONF,
    },
    # Ambernic
    "Win600": {
        "name": "Ambernic Win600",
        "btn_mapping": AMBERNIC_MAPPINGS,
        "extra_buttons": "none",
    },
    # MSI Claw
    "Claw A1M": {
        "name": "MSI Claw (1st gen)",
        # "hrtimer": True, Uses sensor fusion garbage? From intel? Will need custom work
        "extra_buttons": "none",
        "btn_mapping": MSI_CLAW_MAPPINGS,
        "claw": True,
    },
}


def get_default_config(product_name: str, manufacturer: str):
    out = {
        "name": product_name,
        "manufacturer": manufacturer,
        "hrtimer": True,
        "untested": True,
    }

    if manufacturer == "AYA":
        out["btn_mapping"] = AYANEO_BTN_MAPPINGS
        out["mapping"] = AYANEO_DEFAULT_MAPPINGS

    return out



# SPDX-License-Identifier: MIT and GPL-3.0-only
# Sourced from library python-ioctl
# https://github.com/olavmrk/python-ioctl
import ctypes
import platform


class _IoctlGeneric(object):
    _IOC_NRBITS = 8
    _IOC_TYPEBITS = 8
    _IOC_SIZEBITS = 14
    _IOC_DIRBITS = 2
    _IOC_NONE = 0
    _IOC_WRITE = 1
    _IOC_READ = 2

    @classmethod
    def ioc(cls, direction, request_type, request_nr, size):
        _IOC_NRSHIFT = 0
        _IOC_TYPESHIFT = _IOC_NRSHIFT + cls._IOC_NRBITS
        _IOC_SIZESHIFT = _IOC_TYPESHIFT + cls._IOC_TYPEBITS
        _IOC_DIRSHIFT = _IOC_SIZESHIFT + cls._IOC_SIZEBITS
        return (
            (direction << _IOC_DIRSHIFT)
            | (request_type << _IOC_TYPESHIFT)
            | (request_nr << _IOC_NRSHIFT)
            | (size << _IOC_SIZESHIFT)
        )


class _IoctlAlpha(_IoctlGeneric):
    _IOC_NRBITS = 8
    _IOC_TYPEBITS = 8
    _IOC_SIZEBITS = 13
    _IOC_DIRBITS = 3
    _IOC_NONE = 1
    _IOC_READ = 2
    _IOC_WRITE = 4


class _IoctlMips(_IoctlGeneric):
    _IOC_SIZEBITS = 13
    _IOC_DIRBITS = 3
    _IOC_NONE = 1
    _IOC_READ = 2
    _IOC_WRITE = 4


class _IoctlParisc(_IoctlGeneric):
    _IOC_NONE = 0
    _IOC_WRITE = 2
    _IOC_READ = 1


class _IoctlPowerPC(_IoctlGeneric):
    _IOC_SIZEBITS = 13
    _IOC_DIRBITS = 3
    _IOC_NONE = 1
    _IOC_READ = 2
    _IOC_WRITE = 4


class _IoctlSparc(_IoctlGeneric):
    _IOC_NRBITS = 8
    _IOC_TYPEBITS = 8
    _IOC_SIZEBITS = 13
    _IOC_DIRBITS = 3
    _IOC_NONE = 1
    _IOC_READ = 2
    _IOC_WRITE = 4


_machine_ioctl_map = {
    "alpha": _IoctlAlpha,
    "mips": _IoctlMips,
    "mips64": _IoctlMips,
    "parisc": _IoctlParisc,
    "parisc64": _IoctlParisc,
    "ppc": _IoctlPowerPC,
    "ppcle": _IoctlPowerPC,
    "ppc64": _IoctlPowerPC,
    "ppc64le": _IoctlPowerPC,
    "sparc": _IoctlSparc,
    "sparc64": _IoctlSparc,
}


def _machine_ioctl_calculator():
    machine = platform.machine()
    return _machine_ioctl_map.get(machine, _IoctlGeneric)


def _ioc_type_size(size):
    if isinstance(size, type) and issubclass(size, ctypes._SimpleCData):
        return ctypes.sizeof(size)
    elif isinstance(size, int):
        return size
    else:
        raise TypeError(
            "Invalid type for size: {size_type}".format(
                size_type=size.__class__.__name__
            )
        )


def _ioc_request_type(request_type):
    if isinstance(request_type, int):
        return request_type
    if isinstance(request_type, str):
        if len(request_type) > 1:
            raise ValueError("request_type string too long.")
        elif len(request_type) == 0:
            raise ValueError("request_type cannot be an empty string.")
        return ord(request_type)
    else:
        raise ValueError(
            "request_type must be an integer or a string, but was: {request_type_type}".format(
                request_type_type=request_type.__class__.__name__
            )
        )


def _IOC(direction, request_type, request_nr, size):
    """Python implementation of the ``_IOC(...)`` macro from Linux.

    This is a portable implementation of the ``_IOC(...)`` macro from Linux.
    It takes a set of parameters, and calculates a ioctl request number based on those parameters.

    :param direction: Direction of data transfer in this ioctl. This can be one of:

                      * ``None``: No data transfer.
                      * ``'r'``: Read-only (input) data.
                      * ``'w'``: Write-only (output) data.
                      * ``'rw'``: Read-write (input and output) data.
    :param request_type: The ioctl request type. This can be specified as either a string ``'R'`` or an integer ``123``.
    :param request_nr: The ioctl request number. This is an integer.
    :param size: The number of data bytes transferred in this ioctl.
    :return: The calculated ioctl request number.
    """

    calc = _machine_ioctl_calculator()

    if direction is None:
        direction = calc._IOC_NONE
    elif direction == "r":
        direction = calc._IOC_READ
    elif direction == "w":
        direction = calc._IOC_WRITE
    elif direction == "rw":
        direction = calc._IOC_READ | calc._IOC_WRITE
    else:
        raise ValueError("direction must be None, 'r', 'w' or 'rw'.")

    request_type = _ioc_request_type(request_type)
    return calc.ioc(direction, request_type, request_nr, size)


def _IO(request_type, request_nr):
    """Python implementation of the ``_IO(...)`` macro from Linux.

    This is a portable implementation of the ``_IO(...)`` macro from Linux.
    The ``_IO(...)`` macro calculates a ioctl request number for ioctl request that do not transfer any data.

    :param request_type: The ioctl request type. This can be specified as either a string ``'R'`` or an integer ``123``.
    :param request_nr: The ioctl request number. This is an integer.
    :return: The calculated ioctl request number.
    """

    calc = _machine_ioctl_calculator()
    request_type = _ioc_request_type(request_type)
    return calc.ioc(calc._IOC_NONE, request_type, request_nr, 0)


def _IOR(request_type, request_nr, size):
    """Python implementation of the ``_IOR(...)`` macro from Linux.

    This is a portable implementation of the ``_IOR(...)`` macro from Linux.
    The `_IOR(...)`` macro calculates a ioctl request number for ioctl request that only pass read-only (input) data.

    :param request_type: The ioctl request type. This can be specified as either a string ``'R'`` or an integer ``123``.
    :param request_nr: The ioctl request number. This is an integer.
    :param size: The size of the associated data. This can either be an integer or a ctypes type.
    :return: The calculated ioctl request number.
    """

    calc = _machine_ioctl_calculator()
    request_type = _ioc_request_type(request_type)
    size = _ioc_type_size(size)
    return calc.ioc(calc._IOC_READ, request_type, request_nr, size)


def _IOW(request_type, request_nr, size):
    """Python implementation of the ``_IOW(...)`` macro from Linux.

    This is a portable implementation of the ``_IOW(...)`` macro from Linux.
    The ``_IOW(...)`` macro calculates a ioctl request number for ioctl request that only pass write-only (output) data.

    :param request_type: The ioctl request type. This can be specified as either a string ``'R'`` or an integer ``123``.
    :param request_nr: The ioctl request number. This is an integer.
    :param size: The size of the associated data. This can either be an integer or a ctypes type.
    :return: The calculated ioctl request number.
    """

    calc = _machine_ioctl_calculator()
    request_type = _ioc_request_type(request_type)
    size = _ioc_type_size(size)
    return calc.ioc(calc._IOC_WRITE, request_type, request_nr, size)


def _IOWR(request_type, request_nr, size):
    """Python implementation of the ``_IOWR(...)`` macro from Linux.

    This is a portable implementation of the ``_IOWR(...)`` macro from Linux.
    The ``_IOWR(...)`` macro calculates a ioctl request number for ioctl request that use the data for both reading (input) and writing (output).

    :param request_type: The ioctl request type. This can be specified as either a string ``'R'`` or an integer ``123``.
    :param request_nr: The ioctl request number. This is an integer.
    :param size: The size of the associated data. This can either be an integer or a ctypes type.
    :return: The calculated ioctl request number.
    """

    calc = _machine_ioctl_calculator()
    request_type = _ioc_request_type(request_type)
    size = _ioc_type_size(size)
    return calc.ioc(calc._IOC_READ | calc._IOC_WRITE, request_type, request_nr, size)


# struct input_mask {
# 	__u32 type;
# 	__u32 codes_size;
# 	__u64 codes_ptr;
# };
EVIOCGMASK = _IOR("E", 0x92, 4 + 4 + 8)
EVIOCSMASK = _IOW("E", 0x93, 4 + 4 + 8)
# char * is 8
UINPUT_IOCTL_BASE = "U"
UI_SET_UNIQ_STR = lambda l: _IOC("w", UINPUT_IOCTL_BASE, 112, l)
UI_GET_SYSNAME = lambda l: _IOC("r", UINPUT_IOCTL_BASE, 44, l)

# Revokes access to an evdev device
EVIOCREVOKEALL = _IOW("E", 0x94, 4)
JSIOCREVOKEALL = _IOW("j", 0x94, 4)

# Grab clean ioctl
EVIOCGRABCLEAN = _IOW("E", 0x95, 4)

# Hidraw Descriptors
HIDIOCGRDESCSIZE = _IOR("H", 0x01, 4)
HIDIOCGRDESC = _IOR("H", 0x02, 4096 + 4)

__all__ = (
    "_IOC",
    "_IO",
    "_IOR",
    "_IOW",
    "_IOWR",
    "EVIOCSMASK",
    "EVIOCGMASK",
    "EVIOCREVOKEALL",
    "JSIOCREVOKEALL",
    "UI_SET_UNIQ_STR",
)



import logging
import time
from typing import Literal, Sequence

from hhd.controller import Event
from hhd.controller.base import RgbMode, DEBUG_MODE
from hhd.controller.lib.hid import Device

from .const import (
    COMMANDS_GAME,
    COMMANDS_MOUSE,
    RGB_APPLY,
    RGB_INIT,
    RGB_SET,
    WAIT_READY,
    buf,
)

Zone = Literal["all", "left_left", "left_right", "right_left", "right_right"]
GamepadMode = Literal["default", "mouse", "macro"]
Brightness = Literal["off", "low", "medium", "high"]

logger = logging.getLogger(__name__)


def rgb_set_brightness(brightness: Brightness):
    match brightness:
        case "high":
            c = 0x03
        case "medium":
            c = 0x02
        case "low":
            c = 0x01
        case _:
            c = 0x00
    return buf([0x5A, 0xBA, 0xC5, 0xC4, c])


def rgb_command(
    zone: Zone,
    mode: RgbMode,
    direction,
    speed: str,
    red: int,
    green: int,
    blue: int,
    o_red: int,
    o_green: int,
    o_blue: int,
):
    c_direction = 0x00
    set_speed = True

    match mode:
        case "solid":
            # Static
            c_mode = 0x00
            set_speed = False
        case "pulse":
            # Strobing
            # c_mode = 0x0A
            # Spiral is agressive
            # Use breathing instead
            # Breathing
            c_mode = 0x01
            o_red = 0
            o_green = 0
            o_blue = 0
        case "rainbow":
            # Color cycle
            c_mode = 0x02
        case "spiral":
            # Rainbow
            c_mode = 0x03
            red = 0
            green = 0
            blue = 0
            if direction == "left":
                c_direction = 0x01
        case "duality":
            # Breathing
            c_mode = 0x01
        # case "direct":
        #     # Direct/Aura
        #     c_mode = 0xFF
        # Should be used for dualsense emulation/ambilight stuffs
        case _:
            c_mode = 0x00

    c_speed = 0xE1
    if set_speed:
        match speed:
            case "low":
                c_speed = 0xE1
            case "medium":
                c_speed = 0xEB
            case _:  # "high"
                c_speed = 0xF5

    match zone:
        case "left_left":
            c_zone = 0x01
        case "left_right":
            c_zone = 0x02
        case "right_left":
            c_zone = 0x03
        case "right_right":
            c_zone = 0x04
        case _:
            c_zone = 0x00

    return buf(
        [
            0x5A,
            0xB3,
            c_zone,  # zone
            c_mode,  # mode
            red,
            green,
            blue,
            c_speed if mode != "solid" else 0x00,
            c_direction,
            0x00,  # breathing
            o_red,  # these only affect the breathing mode
            o_green,
            o_blue,
        ]
    )


def rgb_set(
    side: str,
    mode: RgbMode,
    direction: str,
    speed: str,
    red: int,
    green: int,
    blue: int,
    red2: int,
    green2: int,
    blue2: int,
):
    match side:
        case "left_left" | "left_right" | "right_left" | "right_right":
            return [
                rgb_command(
                    side, mode, direction, speed, red, green, blue, red2, green2, blue2
                ),
            ]
        case "left":
            return [
                rgb_command(
                    "left_left",
                    mode,
                    direction,
                    speed,
                    red,
                    green,
                    blue,
                    red2,
                    green2,
                    blue2,
                ),
                rgb_command(
                    "left_right",
                    mode,
                    direction,
                    speed,
                    red,
                    green,
                    blue,
                    red2,
                    green2,
                    blue2,
                ),
            ]
        case "right":
            return [
                rgb_command(
                    "right_right",
                    mode,
                    direction,
                    speed,
                    red,
                    green,
                    blue,
                    red2,
                    green2,
                    blue2,
                ),
                rgb_command(
                    "right_left",
                    mode,
                    direction,
                    speed,
                    red,
                    green,
                    blue,
                    red2,
                    green2,
                    blue2,
                ),
            ]
        case _:
            return [
                rgb_command(
                    "all", mode, direction, speed, red, green, blue, red2, green2, blue2
                ),
            ]


INIT_EVERY_S = 10


def process_events(
    events: Sequence[Event],
    prev_mode: str | None,
    global_init=True,
):
    cmds = []
    mode = None
    br_cmd = None
    init = False
    for ev in events:
        if ev["type"] == "led":
            if ev["initialize"]:
                init = True
            if ev["mode"] == "disabled":
                mode = "disabled"
                br_cmd = rgb_set_brightness("off")
                # Certain RGB modes (e.g., rainbow) do not support being set to
                # off. So switch the mode to solid without a color.
                cmds.extend(
                    rgb_set(ev["code"], "solid", "left", "low", 0, 0, 0, 0, 0, 0)
                )
            else:
                match ev["mode"]:
                    case "pulse":
                        mode = "pulse"
                        set_level = False
                    case "rainbow":
                        mode = "rainbow"
                        set_level = True
                    case "duality":
                        mode = "duality"
                        set_level = False
                    case "solid":
                        mode = "solid"
                        set_level = False
                    case "spiral":
                        mode = "spiral"
                        set_level = True
                    case _:
                        assert False, f"Mode '{ev['mode']}' not supported."
                
                if set_level:
                    br_cmd = rgb_set_brightness(ev["brightnessd"])
                else:
                    br_cmd = rgb_set_brightness("high")

                cmds.extend(
                    rgb_set(
                        ev["code"],
                        mode,
                        ev["direction"],
                        ev["speedd"],
                        ev["red"],
                        ev["green"],
                        ev["blue"],
                        ev["red2"],
                        ev["green2"],
                        ev["blue2"],
                    )
                )

    if not mode or (not cmds and mode != "disabled"):
        # Avoid sending init commands without a mode.
        # The exception being the disabled mode, which just sets the led
        # brightness.
        return [], None

    # Set brightness once per update
    if mode != prev_mode:
        init = True
        if not br_cmd:
            br_cmd = rgb_set_brightness("high")

    if br_cmd:
        cmds.insert(0, br_cmd)

    if init:
        # Init should switch modes
        cmds = [
            *cmds,
            RGB_SET,
            RGB_APPLY,
        ]

    # For now, use init since with global init, the gamepad might
    # not initialize properly
    if global_init or init:
        cmds = [
            *RGB_INIT,
            *cmds,
        ]

    return cmds, mode


class RgbCallback:
    def __init__(self) -> None:
        self.prev_mode = None
        self.global_init = True

    def __call__(self, dev: Device, events: Sequence[Event]):
        cmds, mode = process_events(
            events, self.prev_mode, self.global_init
        )
        self.global_init = False
        if mode:
            self.prev_mode = mode
        if not cmds:
            return
        BCK = "\n"
        if DEBUG_MODE:
            logger.warning(
                f"Running RGB commands:\n{BCK.join([cmd[:20].hex() for cmd in cmds])}"
            )
        for r in cmds:
            dev.write(r)


def wait_for_ready(dev: Device, timeout: int = 1):
    # Wait for the device to be ready
    start = time.perf_counter()

    while time.perf_counter() - start < timeout:
        dev.send_feature_report(WAIT_READY)
        # rep = dev.get_feature_report(FEATURE_KBD_APP) # this is not the proper
        # way for this
        # logger.warning(f"Ready: {rep.hex()}")

        # FIXME: Temporary disable since certain allys have issues with it
        return False

        # if rep and rep[0] == 0x5A and rep[2] == 0x0A:
        #     return True
        # else:
        #     time.sleep(0.1)

    logger.error("Ready timeout lapsed.")
    return False


def switch_mode(dev: Device, mode: GamepadMode, kconf={}, first: bool = False):
    match mode:
        case "default":
            cmds = COMMANDS_GAME(kconf)
        # case "macro":
        #     cmds = MODE_MACRO
        case "mouse":
            cmds = COMMANDS_MOUSE(kconf)
        case _:
            assert False, f"Mode '{mode}' not supported."

    for cmd in cmds:
        # wait_for_ready(dev, timeout=5 if first else 1)
        first = False
        # logger.warning(f"Write: {cmd.hex()}")
        dev.write(cmd)



from Xlib.display import Display

from hhd.plugins.overlay.x11 import (
    QamHandlerGamescope,
    get_gamescope_displays,
    get_overlay_display,
    print_debug,
)


def gamescope_debug(args: list[str]):
    if "qam" in args or "menu" in args:
        force_disp = None
        for arg in args:
            if arg.startswith(":"):
                force_disp = arg

        open_menu = "menu" in args
        win = "menu" if open_menu else "QAM"
        print(f"Opening Steam {win}.")
        c = QamHandlerGamescope(force_disp=force_disp, compat_send=False)
        success = c(open_menu)
        c.close()
        if not success:
            import sys

            print(f"Error, could not open {win}.")
            sys.exit(1)
        return

    if not args or not args[0].startswith(":"):
        ds = get_gamescope_displays()
        print(f"Gamescope displays found: {str(ds)}")
        d = get_overlay_display(ds)
        if not d:
            print(f"Overlay display not found, exitting...")
            return
        d, name = d
    else:
        did = args[0]
        name = did
        d = Display(did)

    print(f"Overlay display is '{name}'")

    print("\nDebug Data:")
    print_debug(d, args)
